/*
 * Memory sharing abstractions.
 *
 * Copyright (C) 2018, 2019, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/re/protocols.h>
#include <l4/re/c/rm.h>
#include <l4/sys/err.h>
#include <l4/sys/task.h>
#include <l4/sys/types.h>

#include "cap_alloc.h"
#include "mem_ipc.h"
#include "semaphore.h"

/* Client interfaces. */

#include "dataspace_client.h"
#include "dataspace_factory_client.h"
#include "region_mapper_client.h"


/* Semaphore used if threading has been initialised by ipc_mem_init. */

l4_cap_idx_t ipc_mem_semaphore = L4_INVALID_CAP;



/* Generic memory allocation. */

long ipc_allocate(unsigned long size, void **addr, l4_cap_idx_t *ds)
{
  /* Allocate and attach the memory for the dataspace. */

  if (ipc_new_dataspace(size, 0, 0, ds))
    return -L4_ENOMEM;

  if (ipc_attach_dataspace(*ds, size, addr))
  {
    ipc_cap_free_um(*ds);
    return -L4_ENOMEM;
  }

  return L4_EOK;
}

/* Create a dataspace. Equivalent to l4re_ma_alloc_align. */

long ipc_new_dataspace(long size, unsigned long flags, unsigned long align,
                       l4_cap_idx_t *cap)
{
  return DataspaceFactory_alloc((ref_DataspaceFactory) l4re_env()->mem_alloc,
                                ipc_varg_long(size),
                                ipc_varg_unsigned_long(flags),
                                ipc_varg_unsigned_long(align),
                                cap);
}

/* Attach a dataspace region. Similar to l4re_rm_attach. */

long ipc_attach_dataspace(l4_cap_idx_t cap, unsigned long size, void **addr)
{
  long err;

  err = ipc_semaphore_down(ipc_mem_semaphore);
  if (err)
    return err;

  /* Attach a region starting at or after start/addr, offset is zero. */

  err = RegionMapper_attach((ref_RegionMapper) l4re_env()->rm,
                            (l4_addr_t *) addr, size, L4RE_RM_SEARCH_ADDR, cap,
                            0, L4_PAGESHIFT);

  ipc_semaphore_up(ipc_mem_semaphore);

  return err;
}

/* Detach a dataspace region. Equivalent to l4re_rm_detach_unmap. */

long ipc_detach_dataspace(void *addr)
{
  l4_cap_idx_t ds;
  l4_addr_t start = (l4_addr_t) addr;
  unsigned long size;
  long err;

  err = ipc_semaphore_down(ipc_mem_semaphore);
  if (err)
    return err;

  /* Detach a region of at least size 1, flags 2 (overlap). */

  err = RegionMapper_detach((ref_RegionMapper) l4re_env()->rm, start, 1, 2,
                                 &start, &size, &ds);

  /* Unmap flexpages for the region if successful. */

  if (!err)
    ipc_unmap_region(start, size);

  ipc_semaphore_up(ipc_mem_semaphore);

  return err;
}

/* Obtain a dataspace size. Equivalent to l4re_ds_size. */

long ipc_dataspace_size(l4_cap_idx_t cap, unsigned long *size)
{
  unsigned long flags;

  return Dataspace_info((ref_Dataspace) cap, size, &flags);
}

/* Unmap flexpages in the specified region. */

void ipc_unmap_region(l4_addr_t start, l4_addr_t size)
{
  l4_addr_t limit;
  unsigned int order;

  limit = start + size;

  /* Traverse the region, unmapping flexpages of the appropriate size. */

  while (start < limit)
  {
    order = l4_fpage_max_order(L4_PAGESHIFT, start, start, limit, 0);

    l4_task_unmap(L4RE_THIS_TASK_CAP,
                  l4_fpage(start, order, L4_FPAGE_RWX),
                  L4_FP_ALL_SPACES);

    start += (1UL << order);
  }
}

// vim: tabstop=2 expandtab shiftwidth=2
